(* $Id: rpc_client.ml,v 1.1 1999/04/12 01:23:44 gerd Exp $
 * ----------------------------------------------------------------------
 *
 *)

open Rtypes
open Xdr
open Rpc
open Rpc_common
open Rpc_general
open Rpc_packer
open Unixqueue

exception Message_not_processable


(* The following exceptions are delivered to the callback function: *)

exception Message_lost        
  (* got EOF when some procedure calls were not replied or not even sent *)

exception Message_timeout
  (* After all retransmissions, there was still no reply *)

exception Communication_error of exn
  (* an I/O error happened *)


type connector =
    Inet of (string * int)                        (* Hostname, port *)
  | InetNumeric of (int * int * int * int * int)  (* IP addr, port *)
  | Unix of string                                (* path to unix dom sock *)
  | Descriptor of Unix.file_descr


type call_state = 
    Waiting    (* call has not yet been sent *)
  | Pending    (* call has been sent, still no answer *)
  | Done       (* got answer for the call *)

(* Normally, the state of the call is changed from Waiting to Pending to
 * Done.
 * In the case of a retransmission, the call is added to the waiting calls 
 * again but its state remains 'Pending'.
 *)


type call =
      { mutable proc : string;
	mutable value : packed_value;
	mutable get_result : (unit -> xdr_value) -> unit;

	mutable state : call_state;
	mutable retrans_count : int;        (* retransmission counter *)
	mutable xid : int;

	mutable timeout_id : wait_id option;
	  (* If a timeout handler has been set, this is the corresponding ID *)
      }	


type t =
      { mutable ready : bool;

        mutable trans : Rpc_transport.t;
	mutable prog :  Rpc_program.t;
	mutable prot :  protocol;
        mutable esys :  event_system;
        mutable group : group;

	mutable close : unit -> unit;

	mutable waiting_calls : call Queue.t;
	mutable pending_calls : call list;
	  (* pending_calls: the same call may be added several times to this
	   * list (because of retransmissions)
	   *)

	mutable next_xid : int;

	mutable has_output_resource : bool;
	mutable has_input_resource : bool;

	(* configs: *)
	mutable timeout : float;
        mutable max_retransmissions : int;

	mutable exception_handler : exn -> unit;
      }	


let debug = ref false
	
  (*****)

let authenticate_unix cl =
    failwith "authenticate_unix not yet implemented"

  (*****)

let pass_result cl call f =
  
  (* Stop the timer, if any : *)
  
  begin match call.timeout_id with
    None -> ()
  | Some id -> 
      (* replace the timeout resource by an immediatly reacting resource: *)
      remove_resource cl.esys cl.group (Wait id);
      add_resource cl.esys cl.group (Wait id, 0.0)
  end;

  (* Change the state of the call to 'Done': *)

  call.state <- Done; 

  (* pass 'f' to the call back function: *)

  try
    call.get_result f
  with
    any ->
      begin  (* pass the exception to the exception handler: *)
	try 
	  cl.exception_handler any
	with
	  _ -> ()
      end


let pass_exception cl call x =
  pass_result cl call (fun () -> raise x)


let pass_exception_to_all cl x =
  let fn_list = ref [] in
  let add_fn xid call =
    if not (List.mem_assoc xid !fn_list) then
      fn_list := (xid,call) :: !fn_list
  in
  List.iter (fun call -> add_fn call.xid call) 
            cl.pending_calls;
  Queue.iter (fun call -> add_fn call.xid call) 
             cl.waiting_calls;
  List.iter (fun (xid,call) -> pass_exception cl call x) !fn_list

  (*****)

let rtd = Rpc_transport.descriptor;;  (* abbreviation *)

let add_call cl procname param receiver =
    if not cl.ready then
      failwith "Rpc_client.add_call: client has been shut down";

    let value =
      Rpc_packer.pack_call 
	cl.prog (uint4_of_int cl.next_xid) procname
	"AUTH_NULL" "" "AUTH_NULL" "" param
    in

    let new_call = 
      { proc = procname;
	value = value;
	get_result = receiver;
	state = Waiting;
	retrans_count = cl.max_retransmissions;
	xid = cl.next_xid;
	timeout_id = None;
      }	 in

    Queue.add new_call cl.waiting_calls;
    cl.next_xid <- cl.next_xid + 1;

    if Queue.length cl.waiting_calls > 0 & not cl.has_output_resource then begin
      (* add a resource if there was previously no pending call *)
      add_resource cl.esys cl.group (Wait_out (rtd cl.trans), (-.1.0));
      cl.has_output_resource <- true
    end

  (*****)

let retransmit cl call =
  if call.state = Pending then begin
    if call.retrans_count > 0 then begin
      (* Make the 'call' waiting again *)
      Queue.add call cl.waiting_calls;
      (* Remove retransmission timer *)
      call.timeout_id <- None;
      (* Decrease the retransmission counter *)
      call.retrans_count <- call.retrans_count - 1
    end
    else begin
      (* still no answer after maximum number of retransmissions *)
      cl.pending_calls <- 
	List.filter (fun c -> c.xid <> call.xid) cl.pending_calls;
      (* Remove retransmission timer *)
      call.timeout_id <- None;
      pass_exception cl call Message_timeout;
    end
  end
      

  (*****)

let process_incoming_message cl message =

    (* Got a 'message' for which the corresponding 'call' must be searched: *)

    let xid = 
      try 
	int_of_uint4 (Rpc_packer.peek_xid message) 
      with
	_ -> raise Message_not_processable
	    (* TODO: shut down the connection. This is a serious error *)
    in

    let call = 
      try 
	Listext.first (fun c -> c.xid = xid) cl.pending_calls 
      with
	Not_found -> 
	  (* Strange: Got a message with a session ID that is not pending.
	   * We assume that this is an answer of a very old message that
	   * has been completely timed out.
	   *)
	  raise Message_not_processable
    in

    let result =
      lazy (let (a,b,c,d) = 
	      (* exceptions are passed directly back to the caller *)
	      Rpc_packer.unpack_reply cl.prog call.proc message 
            in d)
    in

    let determine_result () = Lazy.force result in
	    
    (* remove the call from the list of pending calls.  *)

    cl.pending_calls <- List.filter (fun c -> c.xid <> xid) cl.pending_calls;

    (* pass result to the user *)

    pass_result cl call determine_result

  (*****)

let prerr_event ev =
  match ev with
    Input_arrived d -> prerr_endline "Input_arrived"
  | Output_readiness d -> prerr_endline "Output_readiness"
  | Out_of_band d -> prerr_endline "Out_of_band"
  | Timeout (g,Wait_in d) -> prerr_endline "Timeout Wait_in"
  | Timeout (g,Wait_out d) -> prerr_endline "Timeout Wait_out"
  | Timeout (g,Wait_oob d) -> prerr_endline "Timeout Wait_oob"
  | Timeout (g,Wait id) -> prerr_endline "Timeout Wait"
  | Term g  -> prerr_endline "Term"
;;

let event_handler cl esys esys' ev = 
    (* DEBUG *)
    if !debug then prerr_event ev;

    let cl_d = rtd cl.trans in
    match ev with

      (*** event: input data have been arrived ***)

      Input_arrived(_,d) ->
	if d <> cl_d then raise (Equeue.Reject);
	
	(* Check on End-of-file condition: *)

	if Rpc_transport.at_eof cl.trans then begin
	  if !debug then prerr_endline "EH: eof";

	  if cl.has_input_resource then
	    remove_resource cl.esys cl.group (Wait_in cl_d);
	  cl.has_input_resource <- false;

	  if cl.has_output_resource then
	    remove_resource cl.esys cl.group (Wait_out cl_d);
	  cl.has_output_resource <- false;

	  pass_exception_to_all cl Message_lost;
	  cl.close();
	  cl.ready <- false;

	  ()
	end
	else begin

	  (* Receive message parts until there is a complete message *)

	  let again = ref true in

	  while !again do

	    if !debug then prerr_endline "EH: receive_part";

	    (* TODO: catch exceptions and process them *)
	    ignore (Rpc_transport.receive_part cl.trans);
	    
  	    (* Is the message complete? Yes: process it *)

	    if Rpc_transport.is_message_complete cl.trans then begin
	      if !debug then prerr_endline "EH: got message";
	      let message = Rpc_transport.get cl.trans in

	      try
		process_incoming_message cl message
	      with
		Message_not_processable -> 
		  (* DEBUG *)
		  if !debug then prerr_endline "EH: message not processable";
		  ()
	    end;

	    again := not (Rpc_transport.is_buffer_empty cl.trans);

	  done;

	  (* clean up *)

	  Rpc_transport.clean_input cl.trans;

	  (* All calls processed ? *)
	    
	  if cl.pending_calls = [] & cl.has_input_resource then begin
	    if !debug then prerr_endline "EH: done (input)";
	    remove_resource cl.esys cl.group (Wait_in (rtd cl.trans));
	    cl.has_input_resource <- false;
	  end

	    
	end

      (*** event: ready to output data ***)

    | Output_readiness(_,d) ->
	if d <> cl_d then raise (Equeue.Reject);

	if Rpc_transport.is_sending_complete cl.trans then begin
	  (* send next call *)

	  (* DEBUG *)
	  if !debug then prerr_endline "EH: ready to send new message...";

	  let proceed = ref true in

	  while !proceed do
	    try
	      (* try to get the next waiting call from the queue; may raise
	       * Queue.Empty:
	       *)
	      let call = Queue.take cl.waiting_calls in

	      (* If the call is already 'Done', skip it *)

	      if call.state = Done then
		raise Not_found;

  	      (* Change the state of the call. It is now 'pending': *)

	      cl.pending_calls <- call :: cl.pending_calls;
	      call.state <- Pending;

	      (* Add a new resource for the answer: *)

	      if not cl.has_input_resource then
		add_resource cl.esys cl.group (Wait_in cl_d, (-.1.0));
	      cl.has_input_resource <- true;
	    
	      (* Fill the transport container: *)

	      Rpc_transport.put cl.trans call.value;

	      (* DEBUG *)
	      if !debug then prerr_endline "Rpc_transport.put";

	      (* If there should be a timeout handler, add it: *)

	      if cl.timeout > 0.0 then begin
		let id = new_wait_id cl.esys in
		let op = Wait id in
		let handler ues ev e =
		  if e = Timeout(cl.group,op) then begin
		    remove_resource ues cl.group op;
		    retransmit cl call;
		    raise Equeue.Terminate
		  end
		  else raise Equeue.Reject
		in
		add_resource cl.esys cl.group (op, cl.timeout);
		add_handler cl.esys cl.group handler;
		call.timeout_id <- Some id
	      end;

	      proceed := false

	    with Queue.Empty ->
	      (* The queue was empty: Do some cleanups. *)

	      (* DEBUG *)
	      if !debug then prerr_endline "EH: no calls";

	      Rpc_transport.clean_output cl.trans;

	      if cl.has_output_resource then
		remove_resource cl.esys cl.group (Wait_out (rtd cl.trans))
	      else begin
		(* robustness *)
		if exists_resource cl.esys (Wait_out (rtd cl.trans)) then (
		  if !debug then 
		    prerr_endline "Rpc_client: out of sync with internal state";
		  failwith "out of sync";
		)
	      end;
	      cl.has_output_resource <- false;

	      proceed := false

	    | Not_found ->
		(* proceed with the next queue member, if any *)

		(* DEBUG *)
		if !debug then 
		  prerr_endline "EH: found call that has been done";

		()

	  done
	end;

	if not (Rpc_transport.is_sending_complete cl.trans) then begin
	  (* send next part of calling message *)

	  (* DEBUG *)
	  if !debug then prerr_endline "EH: send_part";

	  (* TODO: catch exceptions and process them *)
	  ignore (Rpc_transport.send_part cl.trans);
	  ()
	    (* TODO: catch EPIPE errors and forward them to the receiving
	     * function
	     *)
	end


    | _ -> raise (Equeue.Reject)

  (*****)

let create esys c prot prog =
    (* open connection *)

    let open_inet_socket host port prot =
      let sv_addr = 
    	try 
	  Unix.inet_addr_of_string host
    	with
	  Failure s ->
            try
              let h = Unix.gethostbyname host in
              h.Unix.h_addr_list.(0)
            with
              Not_found ->
            	failwith ("Rpc_client.create: unknown host " ^ host)
      in

      let sock = 
      	Unix.socket
	  Unix.PF_INET
	  (if prot = Tcp then Unix.SOCK_STREAM else Unix.SOCK_DGRAM)
	  0
      in

      try
	Unix.connect sock (Unix.ADDR_INET (sv_addr, port));
	sock
      with
	any ->
	  Unix.close sock;
	  raise any
    in

    let descr =
      match c with
	Inet (host,port) -> 
	  open_inet_socket host port prot
      |	InetNumeric (a,b,c,d,port) -> 
	  open_inet_socket (Printf.sprintf "%d.%d.%d.%d" a b c d) port prot
      |	Unix path ->
	  let s =
      	    Unix.socket
	      Unix.PF_UNIX
	      (if prot = Tcp then Unix.SOCK_STREAM else Unix.SOCK_DGRAM)
	      0
	  in
	  begin try
	    Unix.connect s (Unix.ADDR_UNIX path);
	    s
	  with
	    any ->
	      Unix.close s;
	      raise any
	  end
      |	Descriptor d -> d
    in

    try
      let trans = Rpc_transport.create descr prot BiPipe in

      (* TODO: add_close_action. What should happen if there are unprocessed
       * calls ?
       *)

      let cl = 
	{ ready = true;
	  trans = trans;
      	  prog = prog;
      	  prot = prot;
      	  esys = esys;
	  group = new_group esys;
	  close = (fun () -> ());
      	  waiting_calls = Queue.create();
	  pending_calls = [];
      	  next_xid = 0;
	  has_input_resource = false;
	  has_output_resource = false;
	  timeout = if prot = Udp then 15.0 else (-.1.0);
	  max_retransmissions = 3;
	  exception_handler = (fun _ -> ())
	} 
      in

      let close_connection () =
	begin match c with
	  Descriptor _ -> ()
	| _ -> 
	    begin try
	      cl.ready <- false;
	      (* DEBUG *)
	      if !debug then prerr_endline "Closing connection...";
	      if prot = Tcp then
		Unix.shutdown descr Unix.SHUTDOWN_SEND;
	      Unix.close descr
	    with
	      _ -> ()
	    end
	end
      in

      cl.close <- close_connection;

      let protected_event_handler esys esys' ev =
	try
	  (* DEBUG *)
	  if !debug then prerr_endline "trying event handling...";
	  event_handler cl esys esys' ev;
	  if !debug then prerr_endline "...successful";
	with
	  Equeue.Reject as x -> 
	    (* DEBUG *)
	    if !debug then prerr_endline "...rejecting";
	    raise x
	| Equeue.Terminate as x -> 
	    (* DEBUG *)
	    if !debug then prerr_endline "...terminating";
	    raise x
	| Abort(g,x') as x -> 
	    (* DEBUG *)
	    if !debug then prerr_endline "...aborting";
	    raise x
	| any ->
	    (* DEBUG *)
	    if !debug then 
	      prerr_endline ("EXCEPTION: " ^ Printexc.to_string any);

	    raise (Abort(cl.group, any))
      in
      
      add_handler esys cl.group protected_event_handler;

      (* add an 'abort' handler: *)

      let abort g x =
	(* run through all waiting and pending callbacks and deliver 'x' *)
	(* DEBUG *)
	if !debug then prerr_endline "aborting...";
	pass_exception_to_all cl x;
	close_connection();
	cl.ready <- false;
      in

      add_abort_action esys cl.group abort;

      cl
    with
      any ->
	begin match c with
	  Descriptor _ -> raise any
	| _            -> Unix.close descr; raise any
	end

  (*****)

let configure cl max_retransmission_trials timeout =
  cl.max_retransmissions <- max_retransmission_trials;
  cl.timeout <- timeout


let set_exception_handler cl xh =
  cl.exception_handler <- xh

let shut_down cl =
  if cl.ready then begin
    cl.close();
    pass_exception_to_all cl Message_lost;
    (* DEBUG *)
    if !debug then prerr_endline "clearing client";
    clear cl.esys cl.group
  end


let verbose b =
  debug := b

(* ======================================================================
 * History:
 * 
 * $Log: rpc_client.ml,v $
 * Revision 1.1  1999/04/12 01:23:44  gerd
 * 	Initial revision.
 *
 * 
 *)
