(* $Id: rpc_general.ml,v 1.1 1999/04/04 20:04:56 gerd Exp $ 
 *
 *)


module Listext = struct

(**********************************************************************)

                       (* prefix, suffix *)

(**********************************************************************)

let rec strip_left l n =
  (* remove the first n elements *)
  if n = 0 then
    l
  else if n > 0 then
  ( match l with
    []      -> raise Not_found
  | x :: l' -> strip_left l' (n-1)
  ) else raise Not_found
;;


let rec is_prefix x y =
  (* is x prefix of y? *)
  match x with
    []       -> true
  | x0 :: x' -> ( match y with
                    []       -> false
                  | y0 :: y' -> if x0 = y0 then is_prefix x' y' else false
                )
;;


let suffix l n =
  (* return the last n elements or Not_found *)
  strip_left l (List.length l - n)
;;


let is_suffix x y =
  (* is x suffix of y? *)
  try
    x = suffix y (List.length x)
  with
    Not_found -> false
;;


let rec prefix l n =
  (* return the first n elements or Not_found *)
  if n = 0 then
    []
  else if n > 0 then
  ( match l with
    []      -> raise Not_found
  | x :: l' -> x :: prefix l' (n-1)
  )
  else raise Not_found
;;


let strip_right l n =
  (* remove the last n elements of l *)
  prefix l (List.length l - n)
;;


(**********************************************************************)

               (* string/array functions for lists *)

(**********************************************************************)

let rec sub l i n =
  (* return n elements of l beginning with element i *)
  if i = 0 then
    prefix l n
  else if i > 0 then
  ( match l with
    []      -> raise Not_found
  | x :: l' -> sub l' (i-1) n
  )
  else raise Not_found
;;




(**********************************************************************)

                      (* create special lists *)

(**********************************************************************)

let rec upto a b =
  if a > b then
    []
  else
    a :: (upto (a+1) b)
;;


let rec upto_char a b =
  List.map Char.chr (upto a b)
;;

(**********************************************************************)

                  (* advanced list manipulation *)

(**********************************************************************)

let rec filter f l =
  (* same as flatten(map f l): Each element x is mapped to a list
   * and the element is replaced by that list. Thus you can remove
   * elements or enlarge the number of elements...
   *)
  match l with
    [] -> []
  | x :: l' -> (f x) @ (filter f l')
;;


let delete_all p l =
  (* delete all elements for which p is true *)
  filter (fun x -> if p x then [] else [x]) l
;;


let rec delete_first p l =
  (* delete the first element for which p is true. If there is no element,
   * raise Not_found
   *)
  match l with
    [] -> raise Not_found
  | x::l' -> if p x then l' else x::(delete_first p l')
;;


type alternative = A | B;;

let rec separate2 f l =
  (* put each element into one of two result lists, or drop it *)
  match l with
    []      -> ([], [])
  | x :: l' ->
      let (l1,l2) = separate2 f l' in
      ( match f x with
	  A -> (x::l1, l2)
        | B -> (l1, x::l2)
      )
;;


let delete_dups f l =
  (* f maps the elements of l to their keys. Then all elements of l
   * with duplicate keys are removed.
   * If f is the identity, simply all duplicate elements of l are removed.
   * TODO: use hashtables if l is large
   *)
  let rec rdel keys l =
    match l with
      []      -> []
    | x :: l' ->
	let k = f x in
	if List.mem k keys then
	  rdel keys l'
	else
	  x :: (rdel (k::keys) l')
  in
  rdel [] l
;;

(**********************************************************************)

                         (* search lists *)

(**********************************************************************)


let rec first p l =
  (* return the leftmost member x of l for which p x is true
   * or raise Not_found
   *)
  match l with
    [] -> raise Not_found
  | x :: l' -> if p x then x else first p l'
;;


(**********************************************************************)

                         (* lists as sets *)

(**********************************************************************)

let rec all_distinct l =
  match l with
    []      -> true
  | x :: l' -> not (List.mem x l') & all_distinct l'
;;


let rec sub_set l1 l2 =
  match l1 with
    [] -> true
  | x :: l1' -> (List.mem x l2) & (sub_set l1' l2)
;;


let equal_sets l1 l2 =
  sub_set l1 l2 & sub_set l2 l1
;;

(**********************************************************************)

                       (* associative lists *)

(**********************************************************************)

(* TODO: rename functions from x_assoc to x_alist *)

let rec norm_assoc l =
  (* normalizes l: if (a,b) and (a,b') are both pairs in l and (a,b)
   * comes earlier, remove (a,b')
   *)
  delete_dups fst l
;;


let rec change_assoc l1 l2 =
  (* Iterate over all pairs in l1. If (a,b) is such a pair, lookup a
   * in l2, say (a,b'), and replace (a,b) by (a,b').
   * This means, that the association l1 is changed by the association l2.
   *)
  match l1 with
    []           -> []
  | (a,b) :: l1' ->
      try
      	let b' = List.assoc a l2 in
	(a,b') :: change_assoc l1' l2
      with
	Not_found -> (a,b) :: change_assoc l1' l2
;;


let compose_assoc l1 l2 =
  (* Compute an association list with no duplicates which is the same
   * as l1 @ l2.
   *)
  norm_assoc (l1@l2)
;;


let delete_assoc l keys =
  (* delete keys from l *)
  delete_all (fun (a,b) -> List.mem a keys) l
;;


let rev_assoc x l =
  fst (first (fun (a,b) -> b = x) l)
;;

let rev_mem_assoc x l =
  List.exists (fun (a,b) -> b = x) l
;;

let rev_assq x l =
  fst (first (fun (a,b) -> b == x) l)
;;

end
;;


module Stringext = struct

(**********************************************************************)

     (* list operators for strings read as list of characters *)

(**********************************************************************)

let length s = String.length s ;;
let hd s = if s="" then failwith "hd"
                   else s.[0] ;;
let tl s = if s="" then failwith "tl"
                   else String.sub s 1 (String.length s - 1) ;;
let nth s i = if i < 0 || i >= String.length s then failwith "nth"
                                               else s.[i];;
let rev s =
  let l = String.length s in
  let t = String.create l in
  for i = 0 to (l-1)/2 do
    t.[i] <- s.[l-1-i]
  done;
  t
;;

let concat s_list = List.fold_left (^) "" s_list ;;
let flatten = concat ;;

let iter (f: char->unit) s =
  for i = 0 to String.length s - 1 do
    f (s.[i])
  done;
  ()
;;

let map (f: char->char) s =
  let t = String.copy s in
  for i = 0 to String.length s - 1 do
    t.[i] <- f (s.[i])
  done;
  t
;;


let fold_left (f:'a->char->'a) (x0:'a) s =
  let x = ref x0 in
  for i = 0 to String.length s - 1 do
    x := f !x s.[i]
  done;
  !x
;;


let fold_right (f:char->'a->'a) s (x0:'a) =
  let x = ref x0 in
  for i = String.length s - 1 downto 0 do
    x := f !x s.[i]
  done;
  !x
;;

let for_all f s =
  try
    for i = 0 to String.length s - 1 do
      if not (f (s.[i])) then
	raise Not_found
    done;
    true
  with
    Not_found -> false
;;


let exists f s =
  try
    for i = 0 to String.length s - 1 do
      if f (s.[i]) then
	raise Not_found
    done;
    false
  with
    Not_found -> true
;;


let mem c s = exists (fun a -> a==c) s;;
let memq = mem;;

(* assoc, assq, mem_assoc, split, combine not sensible *)
(* TODO: iter2, map2, fold_left2, fold_right2, for_all2, exists2 *)


let filter f s =
  (* replace characters of s by the string returned by f *)
  let ls = String.length s in
  let rec g i =
    if i < ls then
      f (s.[i]) ^ g (i+1)
    else
      ""
  in g 0
;;


let explode s =
  let l = String.length s in
  let rec r i =
    if i<l then
      s.[i] :: r (i+1)
    else
      []
  in
  r 0
;;


let implode (cl:char list) : string =
  let l = List.length cl in
  let s = String.make l ' ' in
  let rec r cl i =
    match cl with
      [] -> ()
    | c :: cl' ->
	s.[i] <- c;
	r cl' (i+1)
  in
  r cl 0;
  s
;;


(**********************************************************************)

                       (* prefix, suffix *)

(**********************************************************************)

let is_prefix x y =
  (* is x prefix of y? *)
  let lx = String.length x in
  let ly = String.length y in
    lx <= ly && (String.sub y 0 lx = x)
;;


let is_suffix x y =
  (* is x suffix of y? *)
  let lx = String.length x in
  let ly = String.length y in
    lx <= ly && (String.sub y (ly-lx) lx = x)
;;


let strip_left s n =
  (* remove the first n characters *)
  let l = String.length s in
  if n <= n then
    String.sub s n (l-n)
  else
    raise Not_found
;;


let suffix s n =
  (* return the last n elements or Not_found *)
  strip_left s (String.length s - n)
;;


let prefix s n =
  (* return the first n elements *)
  let l = String.length s in
  if n <= n then
    String.sub s 0 n
  else
    raise Not_found
;;


let strip_right s n =
  (* remove the last n elements of s *)
  prefix s (String.length s - n)
;;


(**********************************************************************)

                          (* lists of strings *)

(**********************************************************************)

let rec join slist separator =
  (* concatenate all members of 'slist' and put 'separator' between them *)
  match slist with
    []        -> ""
  | [x]       -> x
  | x::slist' -> x ^ separator ^ (join slist' separator)
;;

end;;

(*======================================================================
 * $Log: rpc_general.ml,v $
 * Revision 1.1  1999/04/04 20:04:56  gerd
 * 	Initial release in this depot. This files have been restructured;
 * some missing features have been added (fp8 code, recursive xdr types).
 *
 *)
