(* $Id: rpc_packer.ml,v 1.1 1999/04/12 01:23:44 gerd Exp $
 * ----------------------------------------------------------------------
 *
 *)

(* packs procedure calls using XDR *)

open Rtypes
open Xdr
open Rpc
open Rpc_common

(* declaration of the RPC message type system: *)

let rpc_ts_unvalidated =
  [ "auth_flavor",         X_enum [ "AUTH_NULL",     int4_of_int 0;
				    "AUTH_UNIX",     int4_of_int 1;
				    "AUTH_SHORT",    int4_of_int 2;
				    "AUTH_DES",      int4_of_int 3 ];
    
    "opaque_auth",         X_struct [ "flavor", X_type "auth_flavor";
				      "body",   X_opaque (uint4_of_int 400) ];
    
    "msg_type",            X_enum [ "CALL",          int4_of_int 0;
				    "REPLY",         int4_of_int 1 ];
    
    "reply_stat",          X_enum [ "MSG_ACCEPTED",  int4_of_int 0;
				    "MSG_DENIED",    int4_of_int 1 ];

    "accept_stat",         X_enum [ "SUCCESS",       int4_of_int 0;
				    "PROG_UNAVAIL",  int4_of_int 1;
				    "PROG_MISMATCH", int4_of_int 2;
				    "PROC_UNAVAIL",  int4_of_int 3;
				    "GARBAGE_ARGS",  int4_of_int 4 ];

    "reject_stat",         X_enum [ "RPC_MISMATCH",  int4_of_int 0;
				    "AUTH_ERROR",    int4_of_int 1 ];
      
    "auth_stat",           X_enum [ "AUTH_BADCRED",       int4_of_int 1;
				    "AUTH_REJECTEDCRED",  int4_of_int 2;
				    "AUTH_BADVERF",       int4_of_int 3;
				    "AUTH_REJECTEDVERF",  int4_of_int 4;
				    "AUTH_TOOWEAK",       int4_of_int 5 ];

    "call_body",           X_struct [ "rpcvers", X_uint;
				      "prog",    X_uint;
				      "vers",    X_uint;
				      "proc",    X_uint;
				      "cred",    X_type "opaque_auth";
				      "verf",    X_type "opaque_auth";
				      "param",   X_param "in" ];

    "call_body_frame",     X_struct [ "rpcvers", X_uint;
				      "prog",    X_uint;
				      "vers",    X_uint;
				      "proc",    X_uint;
				      "cred",    X_type "opaque_auth";
				      "verf",    X_type "opaque_auth";
				      	(* "param",   X_param "in" *) ];

    "accepted_reply",      X_struct [ "verf",    X_type "opaque_auth";
				      "reply_data",
				      X_union_over_enum
					((X_type "accept_stat"),
					 [ "SUCCESS",      X_param "out";
					   "PROG_MISMATCH",
					   X_struct [ "low",  X_uint;
						      "high", X_uint ] ],
					 Some X_void)
				    ];
    
    "rejected_reply",      X_union_over_enum
                             ((X_type "reject_stat"),
			      [ "RPC_MISMATCH",  X_struct [ "low",  X_uint;
							    "high", X_uint ];
				"AUTH_ERROR",    X_type "auth_stat"
			      ],
			      None);
    
    "reply_body",          X_union_over_enum
                             ((X_type "reply_stat"),
			      [ "MSG_ACCEPTED", X_type "accepted_reply";
				"MSG_DENIED",   X_type "rejected_reply" ],
			      None);

    "rpc_msg",             X_struct [ "xid",  X_uint;
				      "body", 
				      X_union_over_enum 
					((X_type "msg_type"),
					 [ "CALL",  X_type "call_body";
					   "REPLY", X_type "reply_body" ],
					 None) ];
      
    "rpc_msg_call_frame",  X_struct [ "xid",  X_uint;
				      "body", 
				      X_union_over_enum 
					((X_type "msg_type"),
					 [ "CALL",  X_type "call_body_frame"
					    (* "REPLY", X_type "reply_body"*) ],
					 None) ];
  ] 

(****)

(* validate message type system on demand and expand the "rpc_msg" symbol,
 * i.e. get the generic type of messages. This type has parameters "in"
 * and "out" for the input and output types of concrete messages, resp.
 *)

let rpc_msg =
  lazy 
    ( let rpc_ts = validate_xdr_type_system rpc_ts_unvalidated in
        expanded_xdr_type rpc_ts (X_type "rpc_msg")
    )

let rpc_msg_call_frame =
  lazy 
    ( let rpc_ts = validate_xdr_type_system rpc_ts_unvalidated in
        expanded_xdr_type rpc_ts (X_type "rpc_msg_call_frame")
    )

let valid_void = validate_xdr_type X_void

(****)

type packed_value = string

(****)
  
let pack_call prog xid proc flav_cred data_cred flav_verf data_verf
              proc_parm =

  let prog_nr = Rpc_program.program_number prog in
  let vers_nr = Rpc_program.version_number prog in
  let proc_nr, in_t, out_t = Rpc_program.signature prog proc in
  
  let message_t = Lazy.force rpc_msg in      (* type of generic message *)

  let message_v =                            (* value of the message *)
    (XV_struct
       [ "xid",  XV_uint xid;
	 "body", XV_union_over_enum
	   ( "CALL", XV_struct 
	              [ "rpcvers", XV_uint (uint4_of_int 2);
			"prog",    XV_uint prog_nr;
			"vers",    XV_uint vers_nr;
			"proc",    XV_uint proc_nr;
			"cred",    XV_struct 
			             [ "flavor", XV_enum flav_cred;
				       "body",   XV_opaque data_cred
				     ];
			"verf",    XV_struct 
			             [ "flavor", XV_enum flav_verf;
				       "body",   XV_opaque data_verf
				     ];
			"param",   proc_parm
		      ]
            )
       ]) in

  pack_xdr_value_as_string 
    message_v            (* the value to pack *)
    message_t            (* the message type... *)
    [ "in", in_t;        (* ...instantiated with input type...*)
      "out", out_t ]     (* ...and output type *)
    
(****)

let unpack_call_frame octets =
  let message_t = Lazy.force rpc_msg_call_frame in
  let message_v =
    try
      unpack_xdr_value octets message_t []
    with 
      Xdr_format_message_too_long v' -> v'
  in
  
  match message_v with
    XV_struct
      [ "xid",  XV_uint xid;
	"body", XV_union_over_enum 
	  ("CALL", 
	   XV_struct
	     [ "rpcvers", XV_uint rpc_version;
	       "prog",    XV_uint prog_nr;
	       "vers",    XV_uint vers_nr;
	       "proc",    XV_uint proc_nr;
	       "cred",    XV_struct 
		            [ "flavor", XV_enum flav_cred;
			      "body",   XV_opaque data_cred
			    ];
	       "verf",    XV_struct 
		            [ "flavor", XV_enum flav_verf;
			      "body",   XV_opaque data_verf
			    ]
	     ])
      ] ->
	if rpc_version = uint4_of_int 2 then
	  xid, prog_nr, vers_nr, proc_nr, 
	  flav_cred, data_cred, flav_verf, data_verf
        else
	    raise (Rpc_cannot_unpack "RPC version not supported")
  | _ ->
      raise (Rpc_cannot_unpack "strange message")
	
(****)

let unpack_call prog proc octets =
  
  let proc_nr, in_t, out_t = Rpc_program.signature prog proc in
  
  let message_t = Lazy.force rpc_msg in      (* type of generic message *)
  
  let message_v =                            (* unpack the value *)
    unpack_xdr_value
      octets                                 (* XDR encoded value *)
      message_t                              (* generic type *)
      [ "in", in_t;                          (* instance for "in" *)
	"out", out_t ]                       (* instance for "out" type *)
  in
  
  match message_v with
    XV_struct
      [ "xid",  XV_uint xid;
	"body", XV_union_over_enum 
	  ("CALL", 
	   XV_struct
	     [ "rpcvers", XV_uint rpc_version;
	       "prog",    XV_uint prog_nr;
	       "vers",    XV_uint vers_nr;
	       "proc",    XV_uint proc_nr;
	       "cred",    XV_struct 
		            [ "flavor", XV_enum flav_cred;
			      "body",   XV_opaque data_cred
			    ];
	       "verf",    XV_struct 
		            [ "flavor", XV_enum flav_verf;
			      "body",   XV_opaque data_verf
			    ];
	       "param",   proc_parm
	     ])
      ] 
    -> if rpc_version = uint4_of_int 2 then
          xid, prog_nr, vers_nr, proc_nr, 
          flav_cred, data_cred, flav_verf, data_verf,
          proc_parm
        else
          raise (Rpc_cannot_unpack "RPC version not supported")
  | _ ->
      raise (Rpc_cannot_unpack "strange message")

(****)

let pack_successful_reply prog proc xid flav_verf data_verf return_value =

  let proc_nr, in_t, out_t = Rpc_program.signature prog proc in

  let message_t = Lazy.force rpc_msg in      (* type of generic message *)

  let message_v =                            (* value of the message *)
    (XV_struct
       [ "xid",  XV_uint xid;
	 "body", XV_union_over_enum 
	   ("REPLY",
	    XV_union_over_enum
	      ("MSG_ACCEPTED",
	       XV_struct
		 [ "verf", XV_struct [ "flavor", XV_enum flav_verf;
				       "body",   XV_opaque data_verf ];
		   "reply_data", XV_union_over_enum
		     ("SUCCESS", return_value)
		 ] ))
       ] ) in
  
  pack_xdr_value_as_string 
    message_v            (* the value to pack *)
    message_t            (* the message type... *)
    [ "in", in_t;        (* ...instantiated with input type...*)
      "out", out_t ]     (* ...and output type *)
 
(****)
      
let pack_accepting_reply xid flav_verf data_verf condition =
  
  let case, explanation = 
    match condition with
      Unavailable_program       -> "PROG_UNAVAIL", XV_void
    | Unavailable_version (l,h) -> "PROG_MISMATCH", 
                                    XV_struct [ "low", XV_uint l;
						"high", XV_uint h ]
    | Unavailable_procedure     -> "PROC_UNAVAIL", XV_void
    | Garbage                   -> "GARBAGE_ARGS", XV_void
    | _                         -> failwith "pack_accepting_reply"
  in

  let message_t = Lazy.force rpc_msg in      (* type of generic message *)

  let message_v =                            (* value of the message *)
    (XV_struct
       [ "xid",  XV_uint xid;
	 "body", XV_union_over_enum 
	   ("REPLY",
	    XV_union_over_enum
	      ("MSG_ACCEPTED",
	       XV_struct
		 [ "verf", XV_struct [ "flavor", XV_enum flav_verf;
				       "body",   XV_opaque data_verf ];
		   "reply_data", XV_union_over_enum
		     (case, explanation)
		 ] ))
       ] ) in
  
  pack_xdr_value_as_string 
    message_v            (* the value to pack *)
    message_t            (* the message type... *)
    [ "in", valid_void;      (* ...instantiated with input type...*)
      "out", valid_void ]    (* ...and output type *)

(****)

let pack_rejecting_reply prog proc xid condition =

  let case, explanation =
    match condition with
      Rpc_mismatch (l,h) -> "RPC_MISMATCH", XV_struct [ "low", XV_uint l;
							"high", XV_uint h ]
    | Auth_bad_cred      -> "AUTH_ERROR", XV_enum "AUTH_BADCRED"
    | Auth_rejected_cred -> "AUTH_ERROR", XV_enum "AUTH_REJECTEDCRED"
    | Auth_bad_verf      -> "AUTH_ERROR", XV_enum "AUTH_BADVERF"
    | Auth_rejected_verf -> "AUTH_ERROR", XV_enum "AUTH_REJECTEDVERF"
    | Auth_too_weak      -> "AUTH_ERROR", XV_enum "AUTH_TOOWEAK"
  in

  let proc_nr, in_t, out_t = Rpc_program.signature prog proc in

  let message_t = Lazy.force rpc_msg in      (* type of generic message *)

  let message_v =                            (* value of the message *)
    (XV_struct
       [ "xid",  XV_uint xid;
	 "body", XV_union_over_enum 
	   ("REPLY",
	    XV_union_over_enum
	      ("MSG_DENIED",
	       XV_union_over_enum (case, explanation)))]) in
  
  pack_xdr_value_as_string 
    message_v            (* the value to pack *)
    message_t            (* the message type... *)
    [ "in", in_t;        (* ...instantiated with input type...*)
      "out", out_t ]     (* ...and output type *)
    
(****)

let unpack_reply prog proc octets =
  
  let proc_nr, in_t, out_t = Rpc_program.signature prog proc in
  
  let message_t = Lazy.force rpc_msg in      (* type of generic message *)
  
  let message_v =                            (* unpack the value *)
    unpack_xdr_value
      octets                                 (* XDR encoded value *)
      message_t                              (* generic type *)
      [ "in", in_t;                          (* instance for "in" *)
	"out", out_t ]                       (* instance for "out" type *)
  in

  try
    let 
      	XV_struct 
        [ "xid",  XV_uint xid;
	  "body", XV_union_over_enum
	            ( "REPLY", XV_union_over_enum
		       ( reply_stat, reply_part1_v )
                    )
        ]
	= message_v
    in
    match reply_stat with
      "MSG_ACCEPTED" ->
	let XV_struct
	    [ "verf",       XV_struct
		              [ "flavor", XV_enum verf_flavor;
				"body",   XV_opaque verf_data
			      ];
	      "reply_data", XV_union_over_enum ( accept_stat, reply_part2_v )
            ]	
	    = reply_part1_v in
	begin
	  match accept_stat with
	    "SUCCESS" ->
	        (* return all what we found: *)
	      (xid, verf_flavor, verf_data, reply_part2_v)
	  | "PROG_UNAVAIL" ->  raise (Rpc_server Unavailable_program)
	  | "PROG_MISMATCH" -> let
		                 XV_struct
		                   [ "low", XV_uint l;
				     "high", XV_uint h ] = reply_part2_v in
	                       raise (Rpc_server (Unavailable_version (l,h)))
	  | "PROC_UNAVAIL" ->  raise (Rpc_server Unavailable_procedure)
	  | "GARBAGE_ARGS" ->  raise (Rpc_server Garbage)
	end
    | "MSG_DENIED"   -> 
	let XV_union_over_enum ( reject_stat, reply_part2_v ) = reply_part1_v in
	begin
	  match reject_stat with
	    "RPC_MISMATCH" ->
	      let XV_struct [ "low", XV_uint l; "high", XV_uint h ] = 
		reply_part2_v in
	      raise (Rpc_server (Rpc_mismatch (l,h)))
	  | "AUTH_ERROR" ->
	      begin
		match reply_part2_v with
		  XV_enum "AUTH_BADCRED" ->
		    raise (Rpc_server Auth_bad_cred)
		| XV_enum "AUTH_REJECTEDCRED" ->
		    raise (Rpc_server Auth_rejected_cred)
		| XV_enum "AUTH_BADVERF" ->
		    raise (Rpc_server Auth_bad_verf)
		| XV_enum "AUTH_REJECTEDVERF" ->
		    raise (Rpc_server Auth_rejected_verf)
		| XV_enum "AUTH_TOOWEAK" ->
		    raise (Rpc_server Auth_too_weak)
	      end
	end
  with
    Match_failure _ as x -> 
      (* raise x; *) (* DEBUG *)
      raise (Rpc_cannot_unpack "Unsupported RPC variant")
	
(****)

let unpack_reply_verifier prog proc octets =
  
  let proc_nr, in_t, out_t = Rpc_program.signature prog proc in
  
  let message_t = Lazy.force rpc_msg in      (* type of generic message *)
    
  let message_v =                            (* unpack the value *)
    unpack_xdr_value
      octets                                 (* XDR encoded value *)
      message_t                              (* generic type *)
      [ "in", in_t;                          (* instance for "in" *)
	"out", out_t ]                       (* instance for "out" type *)
  in

  try
    let 
      	XV_struct 
        [ "xid",  xid;
	  "body", XV_union_over_enum
	           ( "REPLY", XV_union_over_enum
		      ( reply_stat, reply_part1_v )
                   )
        ]
	= message_v
    in
    match reply_stat with
      "MSG_ACCEPTED" ->
	let XV_struct
	    [ "verf",       XV_struct
		              [ "flavor", XV_enum verf_flavor;
				"body",   XV_opaque verf_data
			      ];
	      "reply_data", XV_union_over_enum ( accept_stat, reply_part2_v )
            ]	
	    = reply_part1_v in
	verf_flavor, verf_data
  with
    Match_failure _ as x -> 
      (* raise x; *) (* DEBUG *)
      raise (Rpc_cannot_unpack "Unsupported RPC variant")
	
(*****)

let peek_xid octets =
  if String.length octets < 4 then
    failwith "peek_xid: message too short";
  
  let m4 = String.sub octets 0 4 in
  
  let v = 
    unpack_xdr_value
      m4
      (validate_xdr_type X_uint)
      []
  in

  let XV_uint xid = v in

  xid


let string_of_packed_value s = s;;
let packed_value_of_string s = s;;

(* ======================================================================
 * History:
 * 
 * $Log: rpc_packer.ml,v $
 * Revision 1.1  1999/04/12 01:23:44  gerd
 * 	Initial revision.
 *
 * 
 *)
