(* $Id: rpc_server.ml,v 1.1 1999/04/12 01:23:45 gerd Exp $
 * ----------------------------------------------------------------------
 *
 *)

open Rtypes
open Xdr
open Unixqueue
open Rpc_common
open Rpc

exception Connection_lost

exception Stop_server

type t =
      { mutable master : Unix.file_descr option;        (* master socket *)
	mutable slaves : (Unix.file_descr * connection) list;
	mutable prog : Rpc_program.t;
	mutable procs : (uint4 * binding) list;       (* procnr, binding *)
	mutable esys : event_system;
	mutable group : group;	mutable prot : protocol;
	mutable mode : mode;
	mutable exception_handler : exn -> unit;
	mutable unmap_port : (unit -> unit);
      }	

and connection =
      { whole_server : t;
	mutable ready : bool;
        mutable trans : Rpc_transport.t;

	mutable next_call_id : int;
	
	(* received asynchronous calls: *)
	mutable recv_calls : session list;
	
	(* replies to do: *)
	mutable replies : session list;

	(* RPC does not define how to check if replies are delivered,
	 * so there is no "re-reply" mechanism. The client has to call
	 * again; but the server cannot identify such repetitions.
	 * (The xid field cannot be used for this purpose!)
	 *)

	mutable has_output_resource : bool;
      }	

and session =
      { server : connection;
	peer : Unix.sockaddr option;    (* only used for UDP *)
	call_id : int;
	client_id : uint4;         (* xid *)
	procname : string;
	mutable parameter : xdr_value;     (* XV_void if not used *)
	mutable result : Rpc_packer.packed_value;        
         (* complete result; "" if not used *)
      }	

and connector =
      Localhost of int                     (* port, 0: automatically chosen *)
    | Portmapped      
    | Unix of string                       (* path to unix dom sock *)
    | Descriptor of Unix.file_descr

and binding_sync =
      { sync_name : string;
	sync_proc : xdr_value -> xdr_value
      }	

and binding_async =
      { async_name : string;
	async_invoke : session -> xdr_value -> unit
                                            (* invocation of this procedure *)
      }	
 	    
and binding =
      Sync of binding_sync
    | Async of binding_async


let debug = ref false

  (*****)

let null_packed_value =
  Rpc_packer.packed_value_of_string ""

  (*****)

let process_incoming_message srv conn peer message =

    let new_reply = ref false in

    begin try
      let 
	  xid, prog_nr, vers_nr, proc_nr, 
	  flav_cred, data_cred, flav_verf, data_verf
	  = Rpc_packer.unpack_call_frame message
      in
	      	
	(* TODO: authentication *)

      try
	if prog_nr <> Rpc_program.program_number srv.prog then
	  raise (Rpc_server Unavailable_program);
	let my_vers = Rpc_program.version_number srv.prog in
	if vers_nr <> my_vers then
	  raise (Rpc_server (Unavailable_version (my_vers, my_vers)));
	let proc = 
	  try
	    List.assoc proc_nr srv.procs 
	  with Not_found -> raise (Rpc_server Unavailable_procedure)
	in
		
	(* Wow! We have a procedure to call! *)

	let procname =
	  match proc with 
	    Sync p -> p.sync_name
	  | Async p -> p.async_name
	in
		  
	let 
	    xid', prog_nr', vers_nr', proc_nr', 
	    flav_cred', data_cred', flav_verf', data_verf',
	    param
	    = Rpc_packer.unpack_call srv.prog procname message
	in
		
	match proc with
	  Sync p ->
	    let result_value =
	      (* TODO: exception handling *)
	      p.sync_proc param
	    in
	    let reply = Rpc_packer.pack_successful_reply 
		srv.prog p.sync_name xid "AUTH_NULL" "" result_value in
	    let answer =
	      { server = conn;
		peer = peer;
		call_id = (-1);          (* not applicable *)
		client_id = xid;
		procname = p.sync_name;
		parameter = XV_void;
		result = reply
	      } in
	    conn.replies <- conn.replies @ [answer];
	    new_reply := true

	| Async p ->
	    let this_session =
	      { server = conn;
		peer = peer;
		call_id = conn.next_call_id;
		client_id = xid;
		procname = p.async_name;
		parameter = param;
		result = null_packed_value
	      } in
	    
	    conn.next_call_id <- conn.next_call_id + 1;
	    
	    conn.recv_calls <- conn.recv_calls @ [this_session];
	    p.async_invoke this_session param
	      
      with
	Rpc_server condition -> 
	  let reply = Rpc_packer.pack_accepting_reply xid "AUTH_NULL" "" condition in
	  let answer =
	    { server = conn;
	      peer = peer;
	      call_id = (-1);          (* not applicable *)
	      client_id = xid;
	      procname = "";
	      parameter = XV_void;
	      result = reply
	    } in
	  conn.replies <- conn.replies @ [answer];
	  new_reply := true
			 
    with 
      
      (* TODO: If using Tcp, we SHOULD reply! *)

      Abort(_,_) as x ->
	raise x
    | any ->
	try srv.exception_handler any with _ -> ()
    end;

    !new_reply

  (*****)

let connection_event_handler srv conn esys esys' ev =
    let conn_d = Rpc_transport.descriptor conn.trans in
    match ev with

      (*** event: input data have been arrived ***)

      Input_arrived(_,d) ->
	if d <> conn_d then raise (Equeue.Reject);

	if Rpc_transport.at_eof conn.trans then begin
	  if !debug then prerr_endline "EH: eof";

	  remove_resource srv.esys srv.group (Wait_in conn_d);
	  if conn.has_output_resource then
	    remove_resource srv.esys srv.group (Wait_out conn_d);

          (* TODO: raise error if incomplete message received *)

	  raise Equeue.Terminate
	end
	else begin

	  (* process all what is buffered: *)
	  let again = ref true in
	  while !again do

      	    (* Receive next part of the message *)

	    if !debug then prerr_endline "EH: receive_part";
	    ignore(Rpc_transport.receive_part conn.trans);

	    (* Is the message complete? Yes: process it *)

	    if Rpc_transport.is_message_complete conn.trans then begin
	      if !debug then prerr_endline "EH: got message";
	      let message = Rpc_transport.get conn.trans in
	      let peer = 
	      	if srv.prot = Udp & srv.mode = Socket then begin
		  Some
		    (try Rpc_transport.get_sender conn.trans with
		      Not_found -> failwith "Rpc_server: cannot obtain socket address of peer")
	      	end else None
	      in

	      let new_reply =
		process_incoming_message srv conn peer message
	      in

	      (* ... if there are replies, ensure there is a resource *)

	      if not conn.has_output_resource & new_reply then begin
	      	add_resource srv.esys srv.group (Wait_out d, (-1.0));
		conn.has_output_resource <- true;
	      end

	    end;
	    
	    again := not (Rpc_transport.is_buffer_empty conn.trans);
	  done;

          (* clean up transporter *)

	  Rpc_transport.clean_input conn.trans

	end	  

      (*** event: ready to output data ***)

    | Output_readiness(_,d) ->
	if d <> conn_d then raise (Equeue.Reject);

	if Rpc_transport.is_sending_complete conn.trans then begin
	  (* send next reply *)
	  match conn.replies with
	    next_reply :: other_replies ->
	      if !debug then prerr_endline "EH: next reply";
	      Rpc_transport.put conn.trans next_reply.result;
	      begin match next_reply.peer with
		None -> ()
	      |	Some p -> Rpc_transport.set_receiver conn.trans p
	      end;
	      conn.replies <- other_replies
	  | [] -> 
	    (* this was the last reply in the queue *)
	      if !debug then prerr_endline "EH: last reply";
	      Rpc_transport.clean_output conn.trans;
	      remove_resource srv.esys srv.group (Wait_out d);
	      conn.has_output_resource <- false
	end;

	if not (Rpc_transport.is_sending_complete conn.trans) then begin
	  (* next part of call message *)
	  if !debug then prerr_endline "EH: send_part";
	  ignore(Rpc_transport.send_part conn.trans);
	  ()
	    (* TODO: catch EPIPE errors...
	     *)
	end

    | _ ->
	raise (Equeue.Reject)

  (*****)

let close_conn conn descr =
  if conn.ready then begin
    conn.ready <- false;
    Unix.close descr
  end


let master_event_handler srv master esys esys' ev =
    match ev with
      (*** event: input data have been arrived ***)
      Input_arrived(_,d) ->
	if d <> master then raise (Equeue.Reject);
	(* Note: A master socket can never reach "end of file" *)
	let (slave_d, slave_addr) = Unix.accept d in
	let conn = 
	    { whole_server = srv;
	      ready = true;
	      trans = Rpc_transport.create slave_d srv.prot BiPipe;
	      next_call_id = 0;
	      recv_calls = [];
	      replies = [];
	      has_output_resource = false;
	    } in

	add_handler srv.esys srv.group (connection_event_handler srv conn);
	add_resource srv.esys srv.group (Wait_in slave_d, (-1.0));
	add_close_action 
	  srv.esys 
	  srv.group 
	  (slave_d, (fun d -> close_conn conn d ))
	
    | _ ->
	raise (Equeue.Reject)

  (*****)

let create esys conn prot mode prog procs max_clients =

    let default_exception_handler ex =
      prerr_endline ("Exception " ^ Printexc.to_string ex ^ " caught")
    in

    let srv = 
      { master = None;
	slaves = [];
	prog = prog;
	procs = 
	  List.map
	    (fun b ->
	      let name = 
		match b with 
		  Sync b' -> b'.sync_name 
		| Async b' -> b'.async_name
	      in
	      Rpc_program.procedure_number prog name, b)
	  procs;
	esys = esys;
	group = new_group esys;
	prot = prot;
	mode = mode;
	exception_handler = default_exception_handler;
	unmap_port = (fun () -> ());
      }	
    in

    (* make the socket listening or waiting, whatever is appropriate *)

    let establish_socket s =
      if mode = Socket & prot = Tcp then begin
	(*  make socket listening (assumes that it is a master socket) *)
	Unix.listen s max_clients;
	srv.master <- Some s
      end else begin
	  let c = 
	    { whole_server = srv;
	      ready = true;
	      trans = Rpc_transport.create s prot mode;
	      next_call_id = 0;
	      recv_calls = [];
	      replies = [];
	      has_output_resource = false;
	    } in
	      
	  srv.slaves <- [ s, c ]
      end
    in

    (* auxiliary function to open a socket on 127.0.0.1 *)

    let bind_to_localhost port =
      let s =
      	Unix.socket
	  Unix.PF_INET
	  (if prot = Tcp then Unix.SOCK_STREAM else Unix.SOCK_DGRAM)
	  0
      in
      try
	Unix.bind
	  s 
	  (Unix.ADDR_INET (Unix.inet_addr_of_string "127.0.0.1", port));
	s	  
      with
	any -> Unix.close s; raise any
    in

    (* get the working descriptor, and a flag whether this descriptor should
     * be closed after all has been done
     *)

    let descr, shall_i_close =
      match conn with
	Localhost port ->
	  let s = bind_to_localhost port in
	  s, true
      |	Portmapped ->
	  (* DEBUG *)
	  if !debug then prerr_endline "Portmapped!";
	  let s = bind_to_localhost 0 in
	  begin try
	    let Unix.ADDR_INET(_,port) = Unix.getsockname s in
	    (* DEBUG *)
	    if !debug then prerr_endline ("port: " ^ string_of_int port);
	    let pm = Rpc_portmapper.create_inet "127.0.0.1" in
	    let prog_nr = Rpc_program.program_number prog in
	    let vers_nr = Rpc_program.version_number prog in
	    let old_port = Rpc_portmapper.getport pm prog_nr vers_nr prot in
	    (* DEBUG *)
	    if !debug then 
	      prerr_endline ("old port: " ^ string_of_int old_port);
	    if old_port > 0 then
	      (* remove old binding: *)
	      ignore(Rpc_portmapper.unset pm prog_nr vers_nr prot old_port);
	    (* set new binding: *)
	    let success = Rpc_portmapper.set pm prog_nr vers_nr prot port in
	    Rpc_portmapper.shut_down pm;
	    (* DEBUG *)
	    if !debug then
	      prerr_endline (if success then "success" else "failure");
	    if not success then
	      failwith "Rpc_server: could not register service";
	    srv.unmap_port <-
	      (fun () -> 
		let pm = Rpc_portmapper.create_inet "127.0.0.1" in
		ignore(Rpc_portmapper.unset pm prog_nr vers_nr prot port);
		Rpc_portmapper.shut_down pm);
	    s, true
	  with
	    any -> Unix.close s; raise any
	  end
      |	Unix path ->
	  let s =
      	    Unix.socket
	      Unix.PF_UNIX
	      (if prot = Tcp then Unix.SOCK_STREAM else Unix.SOCK_DGRAM)
	      0
	  in
	  begin try
	    Unix.bind
	      s (Unix.ADDR_UNIX path);
	    s, true
	  with
	    any -> Unix.close s; raise any
	  end
      |	Descriptor d -> d, false
    in

    (* protect the event handlers against exceptions: *)

    let protect_handler h esys esys' ev =
      try
	h esys esys' ev
     with
        Equeue.Reject as x -> raise x
      | Equeue.Terminate as x -> raise x
      | Abort(g,x') as x -> raise x
      | any ->
          (* DEBUG *)
          if !debug then
	    prerr_endline ("EXCEPTION: " ^ Printexc.to_string any);

          raise (Abort(srv.group, any))
    in

    (* the 'abort' handler shut everything down *)

    let abort g x =
      (* If x = Stop_server, do not print debug message *)

      (* DEBUG *)
      if !debug then begin
	if x = Stop_server then 
	  prerr_endline "Stopping server..."
	else 
	  prerr_endline "aborting...";
      end;
      srv.unmap_port();
    in

    (* Now establish the server, add handlers and resources: *)

    try
      establish_socket descr;

      add_abort_action esys srv.group abort;

      begin
	match srv.master with
      	  None -> ()
	| Some m -> 
	    add_handler esys srv.group 
	                        (protect_handler (master_event_handler srv m));
	    add_resource esys srv.group (Wait_in m, (-1.0));
	    if shall_i_close then
	      add_close_action 
	        esys srv.group (m, (fun descr -> Unix.close descr))
      end;
      
      List.iter
	(fun (d,conn) -> 
	  add_handler esys srv.group 
                           (protect_handler(connection_event_handler srv conn));
	  add_resource esys srv.group (Wait_in (Rpc_transport.descriptor conn.trans), (-1.0));
	  if shall_i_close then
	    add_close_action 
	      srv.esys 
	      srv.group 
	      (d, (fun d' -> close_conn conn d' ))
        )
	srv.slaves;
      
      srv
    with
      any ->
	(match conn with Descriptor _ -> () | _ -> Unix.close descr);
	raise any

  (*****)

let get_event_system a_session =
    a_session.server.whole_server.esys

  (*****)

let reply a_session result_value =
    let conn = a_session.server in
    let srv = conn.whole_server in
    if not conn.ready then
      raise Connection_lost;
    let reply = Rpc_packer.pack_successful_reply 
	srv.prog a_session.procname a_session.client_id "AUTH_NULL" "" result_value in
    a_session.parameter <- XV_void;
    a_session.result <- reply;
    conn.replies <- conn.replies @ [a_session];
    conn.recv_calls <-
       List.flatten
	 (List.map
	    (fun call ->
	      if call.call_id = a_session.call_id then [] else [call])
	    conn.recv_calls);
    
    let d = Rpc_transport.descriptor conn.trans in
    if not (exists_resource srv.esys (Wait_out d)) then 
      add_resource srv.esys srv.group (Wait_out d, (-1.0))

let set_exception_handler srv eh =
  srv.exception_handler <- eh

let stop_server srv =
  raise (Abort(srv.group, Stop_server))

let verbose b =
  debug := b


(* ======================================================================
 * History:
 * 
 * $Log: rpc_server.ml,v $
 * Revision 1.1  1999/04/12 01:23:45  gerd
 * 	Initial revision.
 *
 * 
 *)
