(* $Id: rpc_transport.ml,v 1.1 1999/04/12 01:23:46 gerd Exp $
 * ----------------------------------------------------------------------
 *
 *)

(* TODO:
 * - check that all Unix calls are restarted
 *)

open Rtypes
open Xdr
open Rpc
open Rpc_common
open Rpc_general
  
type t =
      { descr : Unix.file_descr;
	prot : protocol;
	mode : mode;

	(* --- additional socket information (esp. UDP) --- *)

	mutable sender : Unix.sockaddr option; (* originator of last message *)
	mutable next_receiver : Unix.sockaddr option; 
	     (* who gets the next message *)
	
	(* --- output --- *)

	mutable out_value : string;      (* includes record mark if needed *)
	mutable out_count : int;

	(* --- input --- *)

	mutable in_rest : string;         (* unprocessed string rest *)
	mutable in_values : string list;  (* received parts in rev order *)
	mutable in_complete : bool;       (* all parts received? *)
	mutable in_eof : bool;

        (* In Tcp mode:
	 * - in_rm:        the current record marker, perhaps incomplete
	 * - in_rm_count:  (0..4) number of valid bytes in in_rm
	 * - in_length:    numeric value of in_rm if complete
	 * - in_last:      true if last fragment
	 * - in_count:     number of bytes read from fragment
	 *)

	mutable in_rm : string;
	mutable in_rm_count : int;
	mutable in_length : int;
	mutable in_last : bool;
	mutable in_count : int
      }	


let debug = ref false

  (****)

let create d p m =
    { descr = d;
      prot = p;
      mode = m;
      sender = None;
      next_receiver = None;
      out_value = "";
      out_count = 0;
      in_rest = "";
      in_values = [];
      in_complete = true;
      in_eof = false;
      in_rm = "";
      in_rm_count = 0;
      in_length = 0;
      in_last = false;
      in_count = 0;
    } 

  (****)

let descriptor t = t.descr

  (****)
 
let set_receiver t r =
    if t.mode = Socket & t.prot = Udp then
      t.next_receiver <- Some r
    else
      failwith "Rpc_transport.set_receiver"

  (****)

let put t pv =
    let v = Rpc_packer.string_of_packed_value pv in
    if String.length t.out_value <> t.out_count then
      failwith "Rpc_transport.put";
    if v = "" then
      failwith "Rpc_transport.put: empty message not allowed";

    let header = match t.prot with
      Udp -> 
	if String.length v > 8000 then
	  failwith "Rpc_transport.put: UDP message too large";
	""
    | Tcp ->         
	let l = String.length v in
        let l_uint4 = uint4_of_int l in
        let (c3,c2,c1,c0) = dest_uint4 l_uint4 in
        let c3' = Char.chr (Char.code c3 lor 0x80) in
        Stringext.implode [c3';c2;c1;c0] 
    in

    t.out_value <- header ^ v;
    t.out_count <- 0

  (****)

let send_part t =
    let l = String.length t.out_value in
    if l <> t.out_count then begin
      ignore(Unix.select [] [t.descr] [] (-1.0));  
        (* block until writing possible *)
      match t.prot with
	Udp -> 
	  let n =
	    unix_call
	      (fun () -> 
		match t.next_receiver with
		  None -> Unix.write t.descr t.out_value 0 l
		| Some a -> Unix.sendto t.descr t.out_value 0 l [] a) in
	  if n < l then
	    failwith "Rpc_transport.send_part: UDP message too large";
	  t.out_count <- l;
	  true
      |	Tcp ->
	  let n =
	    unix_call
	      (fun () -> Unix.write 
		           t.descr 
		           t.out_value 
		           t.out_count 
		           (l - t.out_count)) in
	  t.out_count <- t.out_count + n;
	  t.out_count = l
    end
    else
      true

  (****)

let is_sending_complete t =
    String.length t.out_value = t.out_count

  (****)

let clean_output t =
    t.out_value <- "";
    t.out_count <- 0

  (****)

let send_sync t pv =
    put t pv;
    try
      while not (is_sending_complete t) do
	ignore(send_part t)
      done;
      clean_output t
    with
      any ->
	clean_output t;
	raise any

  (****)

let recv_buffer = ref ""
let recv_buffer_size = 8000

let receive_part t =
    if t.in_complete then begin
      t.in_complete <- false;
      t.in_values <- [];
      t.in_rm <- "";
      t.in_rm_count <- 0;
      t.in_length <- 0;
      t.in_last <- false;
      t.in_count <- 0
    end;

    if t.in_rest = "" then begin
      ignore(Unix.select [t.descr] [] [] (-1.0));  
        (* block until reading possible *)
      ()
    end;

    if !recv_buffer = "" then
      recv_buffer := String.make recv_buffer_size ' ';

    match t.prot with
      Udp ->
        let n,sender = 
	  match t.mode with
	    Socket ->
	      let n, s =
	      	unix_call 
	    	  (fun () -> Unix.recvfrom 
		               t.descr !recv_buffer 0 recv_buffer_size [])
	      in
	        n, Some s
	  | BiPipe ->
	      unix_call 
	    	(fun () -> Unix.read 
		             t.descr !recv_buffer 0 recv_buffer_size),
	      None
	in
	if n = 0 then begin
	  t.sender <- sender;
	  t.in_eof <- true;
	  false
	end else
	begin
	  t.sender <- sender;
          t.in_values <- [String.sub !recv_buffer 0 n];
	  t.in_complete <- true;
	  true
	end
    | Tcp ->
        let n = 
	  if t.in_rest = "" then
	    unix_call 
	      (fun () -> Unix.read t.descr !recv_buffer 0 8000) 
	  else begin
	    String.blit 
	      t.in_rest 0 !recv_buffer 0 (String.length t.in_rest);
	    String.length t.in_rest
	  end in

	if !debug then prerr_endline ("len=" ^ string_of_int n);
	    
	t.sender <- None;
	t.in_rest <- "";

	if n = 0 then begin
	  t.in_eof <- true;
	  false
	end
	else begin

	  let k = ref 0 in
	  while !k < n  &  not t.in_complete do
  	    (* case: read record marker *)
	    if t.in_rm_count < 4 then begin
	      if !debug then
		prerr_endline ("Record marker k = " ^ string_of_int !k);
	      let m = min (4 - t.in_rm_count) (n - !k) in
	      t.in_rm <- t.in_rm ^ (String.sub !recv_buffer !k m);
	      t.in_rm_count <- t.in_rm_count + m;
	      k := !k + m;
	      (* case: record marker has become complete *)
	      if t.in_rm_count = 4 then begin
	      	let rm = String.copy t.in_rm in
	      	t.in_last <- (Char.code rm.[0]) >= 128;
		if t.in_last & !debug then
		  prerr_endline ("Recognized Last Fragment");
	      	rm.[0] <- (Char.chr ((Char.code rm.[0]) land 0x7f));
	      	t.in_length <- 
		   int_of_uint4 
		     (mk_uint4 (rm.[0],rm.[1],rm.[2],rm.[3]));
	      	t.in_count <- 0
	      end
	    end
	    (* case: read fragment *)
	    else begin
	      if !debug then prerr_endline ("fragment k = " ^ string_of_int !k);
	      let m = min (t.in_length - t.in_count) (n - !k) in
	      t.in_values <- 
		 (String.sub !recv_buffer !k m) :: t.in_values;
	      k := !k + m;
	      t.in_count <- t.in_count + m;
	      (* case: fragment complete *)
	      if t.in_count = t.in_length then begin
	      	if !debug then
		  prerr_endline ("fragment complete k = " ^ string_of_int !k);
	        (* case: last fragment complete *)
	      	if t.in_last then begin
		  if !debug then prerr_endline ("last fragment");
		  t.in_rest <- String.sub !recv_buffer !k (n - !k);
		  k := n;
		  t.in_complete <- true
	      	end;
	      	t.in_rm <- "";
	      	t.in_rm_count <- 0;
	      end
	    end
	  done;

	  if !debug then
	    prerr_endline ("rest n = " ^ string_of_int (String.length t.in_rest));

	  t.in_complete
	end

  (****)

let get_sender t =
    match t.sender with
      Some x -> x
    | None   -> raise Not_found

  (****)

let is_message_complete t =
    t.in_complete & t.in_values <> [] 

  (****)

let at_eof t =
    t.in_eof

  (****)

let is_buffer_empty t =
    t.in_rest = ""

  (****)

let get t =
    if is_message_complete t then
      Rpc_packer.packed_value_of_string
	(String.concat "" (List.rev t.in_values))
    else
      failwith "Rpc_transport.get: message not yet complete"

  (****)

let clean_input t =
    if t.in_complete or t.in_eof then
      t.in_values <- []
    else
      failwith "Rpc_transport.clean_input: message not yet complete"

  (****)

  (* TODO: 
   * - EOF detection
   * - buffer_empty condition
   *)

let receive_sync t =
    if not t.in_complete then
      failwith "Rpc_transport.receive_sync: last message not complete";
    
    let complete = ref false in

    while not !complete do
      complete := receive_part t;
      if at_eof t then raise End_of_file
    done;

    let v = get t in

    clean_input t;

    v


let verbose b =
  debug := b

(* ======================================================================
 * History:
 * 
 * $Log: rpc_transport.ml,v $
 * Revision 1.1  1999/04/12 01:23:46  gerd
 * 	Initial revision.
 *
 * 
 *)
