(*
 * $Id: rtypes.mli,v 1.1 1999/04/04 20:04:56 gerd Exp $
 *)

(* 1-, 2-, 4- and 8-bytes representation of integers,
 * represented as 2's complements
 *)

type int1
type int2
type int4
type int8

(* 1-, 2-, 4- and 8-bytes representation of non-negative integers *)

type uint1
type uint2
type uint4
type uint8

(* Floating-point numbers of single and double precision according to IEEE *)

type fp4
type fp8

(* raised if a conversion can't be done *)

exception Cannot_represent

(* create integers from character tuples. In these tuples, the MSB
 * comes first
 *)

val mk_int1 : char -> int1
val mk_int2 : char * char -> int2
val mk_int4 : char * char * char * char -> int4
val mk_int8 : char * char * char * char * char * char * char * char -> int8
val mk_uint1 : char -> uint1
val mk_uint2 : char * char -> uint2
val mk_uint4 : char * char * char * char -> uint4
val mk_uint8 : char * char * char * char * char * char * char * char -> uint8

(* destroy integers and get tuples *)

val dest_int1 : int1 -> char
val dest_int2 : int2 -> char * char
val dest_int4 : int4 -> char * char * char * char
val dest_int8 : int8 -> char * char * char * char * char * char * char * char
val dest_uint1 : uint1 -> char
val dest_uint2 : uint2 -> char * char
val dest_uint4 : uint4 -> char * char * char * char
val dest_uint8 : uint8 -> char * char * char * char * char * char * char * char

(* integers as XDR compatible strings (i.e. big endian) *)

val int1_as_string : int1 -> string
val int2_as_string : int2 -> string
val int4_as_string : int4 -> string
val int8_as_string : int8 -> string
val uint1_as_string : uint1 -> string
val uint2_as_string : uint2 -> string
val uint4_as_string : uint4 -> string
val uint8_as_string : uint8 -> string


(* conversions from int to [u]intn and vice versa.
 * On 32-bit computers, the type int can hold 31-bit signed integers
 * (including the sign, i.e. one bit cannot be used).
 * On 64-bit computers, the type int can hold 63-bit signed integers
 * (including the sign, i.e. one bit cannot be used).
 * The int_of_xxx functions raise Cannot_represent if the number to
 * convert is too big (or too small) to be represented as int. Note
 * that this depends on the word size of your architecture.
 *)

val int_of_int1 : int1 -> int
val int_of_uint1 : uint1 -> int
val int_of_int2 : int2 -> int
val int_of_uint2 : uint2 -> int
val int_of_int4 : int4 -> int
val int_of_uint4 : uint4 -> int
val int_of_int8 : int8 -> int
val int_of_uint8 : uint8 -> int

val int1_of_int : int -> int1
val uint1_of_int : int -> uint1
val int2_of_int : int -> int2
val uint2_of_int : int -> uint2
val int4_of_int : int -> int4
val uint4_of_int : int -> uint4
val int8_of_int : int -> int8
val uint8_of_int : int -> uint8

(* Note: The "fp8" code is implemented but in a very O'Caml-dependend way.
 * "fp4" code does not exist yet.
 *)

val float_of_fp4 : fp4 -> float
val float_of_fp8 : fp8 -> float
val fp4_of_float : float -> fp4
val fp8_of_float : float -> fp8

(* mk_fp4 e m2 m1 m0:
 * e: sign + exponent
 * m2, m1, m0: normalized mantissa
 *
 * mk_fp8 e1 e0m6 m5 m4 m3 m2 m1 m0:
 * e1: sign + upper 7 bits of exponent
 * e0m6: lower 4 bits of exponent, highest 4 bits of mantissa
 * m5 to m0: lower bits of mantissa
 *)

val mk_fp4 : char * char * char * char -> fp4
val mk_fp8 : char * char * char * char * char * char * char * char -> fp8
val dest_fp4 : fp4 -> char * char * char * char
val dest_fp8 : fp8 -> char * char * char * char * char * char * char * char

val fp4_as_string : fp4 -> string
val fp8_as_string : fp8 -> string

(* ======================================================================
 *
 * $Log: rtypes.mli,v $
 * Revision 1.1  1999/04/04 20:04:56  gerd
 * 	Initial release in this depot. This files have been restructured;
 * some missing features have been added (fp8 code, recursive xdr types).
 *
 *)
