(*
 * $Id: xdr.ml,v 1.3 1999/04/11 22:16:12 gerd Exp $
 *)

(* This is an XDR implementation.
 * See RFC 1014
 *)

open Rtypes;;
open Rpc_general;;

exception Propagate of string;;

(**********************************************************************)
(* definition of XDR types and type systems                           *)
(**********************************************************************)

(* restriction: it is not allowed to have an X_param as enumerator type
 * in union_over_enum. There must always be a real X_enum or, in a 
 * type system, a resolvable X_type at this position.
 *)


type xdr_type_term =
    X_int
  | X_uint
  | X_hyper
  | X_uhyper
  | X_enum of (string * int4) list
  | X_float
  | X_double
  | X_opaque_fixed of uint4
  | X_opaque of uint4
  | X_string of uint4
  | X_array_fixed of xdr_type_term * uint4
  | X_array of       xdr_type_term * uint4
  | X_struct of (string * xdr_type_term) list
  | X_union_over_int of   
      (int4   * xdr_type_term) list * xdr_type_term option
  | X_union_over_uint of  
      (uint4  * xdr_type_term) list * xdr_type_term option
  | X_union_over_enum of  
      xdr_type_term * (string * xdr_type_term) list * xdr_type_term option
  | X_void
  | X_type of string
  | X_param of string
  | X_rec of (string * xdr_type_term)      (* define a recursive type *)
  | X_refer of string                      (* refer to a recursive type *)
;;


type xdr_type = 
  { mutable term   : xdr_term;
    mutable params : string list
      (* "params" is normally only non-empty in the top node *)
  } 
and xdr_term =
    T_int
  | T_uint
  | T_hyper
  | T_uhyper
  | T_enum of (string * int4) list
  | T_float
  | T_double
  | T_opaque_fixed of uint4
  | T_opaque of uint4
  | T_string of uint4
  | T_array_fixed of xdr_type * uint4
  | T_array of       xdr_type * uint4
  | T_struct of (string * xdr_type) list
  | T_union_over_int of
      (int4 * xdr_type) list * xdr_type option
  | T_union_over_uint of  
      (uint4 * xdr_type) list * xdr_type option
  | T_union_over_enum of  
      xdr_type * (string * xdr_type) list * xdr_type option
  | T_void
  | T_param of string
  | T_rec of (string * xdr_type)
  | T_refer of (string * xdr_type)
;;

type xdr_type_term_system =
  (string * xdr_type_term) list
;;

type xdr_type_system = 
  (string * xdr_type) list
  (* export xdr_type_system in an opaque manner *)



let x_bool =
  X_enum ["FALSE", int4_of_int 0; "TRUE", int4_of_int 1]
;;


let x_optional t =
  X_union_over_enum 
    (x_bool,
     ["TRUE", t; "FALSE", X_void],
     None)
;;


let x_opaque_max =
  X_opaque (mk_uint4 ('\063', '\255', '\255', '\255'));;

let x_string_max =
  X_string (mk_uint4 ('\063', '\255', '\255', '\255'));;

let x_array_max t =
  X_array (t,  (mk_uint4 ('\063', '\255', '\255', '\255')));;

(**********************************************************************)
(* definition of XDR values                                           *)
(**********************************************************************)
				     
type xdr_value = 
    XV_int of int4
  | XV_uint of uint4
  | XV_hyper of int8
  | XV_uhyper of uint8
  | XV_enum of string
  | XV_float of fp4
  | XV_double of fp8
  | XV_opaque of string
  | XV_string of string
  | XV_array of xdr_value list
  | XV_struct of (string * xdr_value) list
  | XV_union_over_int of (int4 * xdr_value)
  | XV_union_over_uint of (uint4 * xdr_value)
  | XV_union_over_enum of (string * xdr_value)
  | XV_void
;;

let xv_true = XV_enum "TRUE";;
let xv_false = XV_enum "FALSE";;

let xv_none = XV_union_over_enum ("FALSE",XV_void);;
let xv_some v = XV_union_over_enum ("TRUE",v);;

let dest_xv_int v =    match v with XV_int x -> x 
                                         | _ -> failwith "dest_xv_int";;
let dest_xv_uint v =   match v with XV_uint x -> x 
                                          | _ -> failwith "dest_xv_uint";;
let dest_xv_hyper v =  match v with XV_hyper x -> x 
                                           | _ -> failwith "dest_xv_hyper";;
let dest_xv_uhyper v = match v with XV_uhyper x -> x 
                                            | _ -> failwith "dest_xv_uhyper";;
let dest_xv_enum v =   match v with XV_enum x -> x 
                                          | _ -> failwith "dest_xv_enum";;
let dest_xv_float v =  match v with XV_float x -> x 
                                           | _ -> failwith "dest_xv_float";;
let dest_xv_double v = match v with XV_double x -> x 
                                            | _ -> failwith "dest_xv_double";;
let dest_xv_opaque v = match v with XV_opaque x -> x 
                                            | _ -> failwith "dest_xv_opaque";;
let dest_xv_string v = match v with XV_string x -> x 
                                            | _ -> failwith "dest_xv_string";;
let dest_xv_array v =  match v with XV_array x -> x 
                                           | _ -> failwith "dest_xv_array";;
let dest_xv_struct v = match v with XV_struct x -> x 
                                            | _ -> failwith "dest_xv_struct";;
let dest_xv_void v =   match v with XV_void -> () 
                                        | _ -> failwith "dest_xv_void";;

let dest_xv_union_over_int v = 
  match v with XV_union_over_int x -> x 
                               | _ -> failwith "dest_xv_union_over_int";;
let dest_xv_union_over_uint v = 
  match v with XV_union_over_uint x -> x 
                                | _ -> failwith "dest_xv_union_over_uint";;
let dest_xv_union_over_enum v = 
  match v with XV_union_over_enum x -> x 
                                | _ -> failwith "dest_xv_union_over_enum";;


exception Xdr_format of string;;
exception Xdr_format_message_too_long of xdr_value;;
(* raised in unpack_xdr_value if the byte stream does not match
 * the expected type. The string is an explanation and might be
 * useful while debugging. In the special case Xdr_format_message_too_long
 * there are more bytes than expected, but a prefix matches the type.
 * The prefix is returned as xdr_value.
 *)

(**********************************************************************)
(* check if XDR types are well-formed                                 *)
(**********************************************************************)

(* TODO: check on recursions without finite fix point. *)


let rec validate_xdr_type_i1 
        (r:xdr_type_term -> xdr_type) 
        (b:(string * xdr_type) list)
        (t:xdr_type_term) 
      : xdr_type =

  (* r: function that resolves X_type references
   * t: the xdr_type_term to validate
   * b: list of recursive bindings
   *
   * raise Not_found on any error
   *)

  let mktype tm = { term = tm; params = [] } in

  match t with
    X_int    -> mktype T_int
  | X_uint   -> mktype T_uint
  | X_hyper  -> mktype T_hyper
  | X_uhyper -> mktype T_uhyper
  | X_float  -> mktype T_float
  | X_double -> mktype T_double
  | X_void   -> mktype T_void
  | X_enum e ->
      let e_names, e_values = List.split e in
      if Listext.all_distinct e_names & Listext.all_distinct e_values then
	mktype (T_enum e)
      else
	raise (Propagate "Bad enumeration type: double values")
  | X_opaque_fixed n    -> mktype (T_opaque_fixed n)
  | X_opaque n          -> mktype (T_opaque n)
  | X_string n          -> mktype (T_string n)
  | X_array_fixed (s,n) -> mktype (T_array_fixed(validate_xdr_type_i1 r b s, n))
  | X_array (s,n)       -> mktype (T_array      (validate_xdr_type_i1 r b s, n))
  | X_struct s ->
      let s_names, s_types = List.split s in
      if Listext.all_distinct s_names then
	mktype
	  (T_struct 
	     (List.map (fun (n,x) -> n,validate_xdr_type_i1 r b x) s))
      else
	raise (Propagate "Bad struct type: components with same names found")
  | X_union_over_int (u, default) ->
      let u_values, u_types = List.split u in
      if Listext.all_distinct u_values then begin
	let default' =
	  match default with
	    Some d -> Some (validate_xdr_type_i1 r b d)
	  | None   -> None
	in
	mktype 
	  (T_union_over_int
	     (List.map (fun (n,x) -> n,validate_xdr_type_i1 r b x) u, default'))
      end
      else
	raise (Propagate "Bad union_over_int type: variants found with same tags")
  | X_union_over_uint (u,default) ->
      let u_values, u_types = List.split u in
      if Listext.all_distinct u_values then begin
	let default' =
	  match default with
	    Some d -> Some (validate_xdr_type_i1 r b d)
	  | None   -> None
	in
	mktype 
	  (T_union_over_uint
	     (List.map (fun (n,x) -> n,validate_xdr_type_i1 r b x) u, default'))
      end
      else
	raise (Propagate "Bad union_over_uint type: variants found with same tags")
  | X_union_over_enum (e,u,default) ->
      let e' = validate_xdr_type_i1 r b e in
      let u_values, u_types = List.split u in
      let el =
	match e'.term with
	  T_enum x -> x
	| _ -> raise (Propagate "Bad union_over_enum type: discriminator is not enumerator")
      in
      let el_names, el_values = List.split el in
      if Listext.all_distinct u_values &
	 List.for_all (fun n -> List.mem n el_names) u_values then begin
	   let default' =
	     match default with
	       Some d -> Some (validate_xdr_type_i1 r b d)
	     | None   -> None
	   in
	   mktype 
	     (T_union_over_enum
		(e', 
		 List.map (fun (n,x) -> n,validate_xdr_type_i1 r b x) u,
		 default'))
	 end
      else
	raise (Propagate "Bad union_over_enum type: variants found with same tags")
  | X_type _ ->
      r t
  | X_param p ->
      mktype (T_param p)
  | X_rec (name, s) ->
      let node = mktype T_void in
      let t' = validate_xdr_type_i1 r ((name,node)::b) s in
      node.term <- T_rec (name, t');
      node

  | X_refer name ->
      mktype (T_refer (name, List.assoc name b))
;;


let rec find_params (t:xdr_type) : string list =
  (* collect all parameters *)
  match t.term with
    T_param p -> 
      [p]
  | T_array_fixed (t',n) -> 
      find_params t'
  | T_array (t',n) -> 
      find_params t'
  | T_struct l ->
      Listext.filter (fun (s,t') -> find_params t') l
  | T_union_over_int (l,None) ->
      Listext.filter (fun (n,t') -> find_params t') l
  | T_union_over_int (l,Some u) ->
      Listext.filter (fun (n,t') -> find_params t') l @ find_params u
  | T_union_over_uint (l,None) ->
      Listext.filter (fun (n,t') -> find_params t') l
  | T_union_over_uint (l,Some u) ->
      Listext.filter (fun (n,t') -> find_params t') l @ find_params u
  | T_union_over_enum (e,l,None) ->
      Listext.filter (fun (n,t') -> find_params t') l
  | T_union_over_enum (e,l,Some u) ->
      Listext.filter (fun (n,t') -> find_params t') l @ find_params u
  | T_rec (_,t') ->
      find_params t'
  | _ ->
      []
;;


let rec validate_xdr_type (t:xdr_type_term) : xdr_type =
  let r n =
    raise (Propagate "Cannot resolve X_type element")
  in
  try
    let t' = validate_xdr_type_i1 r [] t in
    let pl = Listext.delete_dups (fun x->x) (find_params t') in
    t'.params <- pl;
    t'
  with
    Not_found ->
      failwith "validate_xdr_type: unspecified error"
  | Propagate s ->
      failwith ("validate_xdr_type: " ^ s)
;;


let rec expand_X_type (s:xdr_type_system) (t:xdr_type_term) : xdr_type =
  match t with
    X_type n ->
      begin
	let rec r s1 s2 =
	  match s2 with
	    []       -> raise (Propagate ("Cannot resolve X_type " ^ n))
	  | (n',t') :: s2' -> 
	      if n = n' then
		t'
	      else
		r (s1 @ [n',t']) s2'
	in
	r [] s
      end
  | _ ->
      raise (Propagate "Found X_type where it must not occur")
;;


let validate_xdr_type_system (s:xdr_type_term_system) : xdr_type_system =
  let names = List.map fst s in
  if Listext.all_distinct names then begin
    let rec r (s1:xdr_type_system) (s2:xdr_type_term_system) =
      match s2 with
	[]           -> []
      |	(n,t) :: s2' -> 
	  let t2 =
	  begin
	    try 
	      let t' = validate_xdr_type_i1 (expand_X_type s1) [] t in
	      let pl = Listext.delete_dups (fun x->x) (find_params t') in
	      t'.params <- pl;
	      t'
	    with
	      Not_found -> failwith "validate_xdr_type_system: unspecified error"
	    | Propagate s -> failwith ("validate_xdr_type_system: " ^ s)
	  end
	  in
	  (n,t2)::(r (s1 @ [n,t2]) s2')
    in
    r [] s
  end
  else
    failwith "validate_xdr_type_system: type system has members with same names"
;;


(**********************************************************************)
(* the reverse way                                                    *)
(**********************************************************************)


let rec xdr_type_term (t:xdr_type) : xdr_type_term =
  let conv_list l =
    List.map (fun (x, t') -> x, xdr_type_term t') l in
  let conv_option p =
    match p with None -> None | Some t' -> Some (xdr_type_term t') in

  match t.term with
    T_int    -> X_int
  | T_uint   -> X_uint
  | T_hyper  -> X_hyper
  | T_uhyper -> X_uhyper
  | T_enum l -> X_enum l
  | T_float  -> X_float
  | T_double -> X_double
  | T_void   -> X_void
  | T_param p        -> X_param p
  | T_opaque_fixed n -> X_opaque_fixed n
  | T_opaque n       -> X_opaque n
  | T_string n       -> X_string n
  | T_array_fixed (t', n) -> X_array_fixed (xdr_type_term t',n)
  | T_array (t', n)       -> X_array       (xdr_type_term t',n)
  | T_struct s       -> X_struct (conv_list s)
  | T_rec (n, t')    -> X_rec (n, xdr_type_term t')
  | T_refer (n, t')  -> X_refer n
  | T_union_over_int (u,d)  -> X_union_over_int  (conv_list u, conv_option d)
  | T_union_over_uint (u,d) -> X_union_over_uint (conv_list u, conv_option d)
  | T_union_over_enum (e,u,d) -> 
      X_union_over_enum (xdr_type_term e, conv_list u, conv_option d)
;;


let xdr_type_term_system (s:xdr_type_system) : xdr_type_term_system = 
  List.map (fun (n,t) -> n,xdr_type_term t) s
;;


(**********************************************************************)
(* expand X_type members relative to given systems                    *)
(**********************************************************************)

(* The implemantation of "expanded_xdr_type_term" repeats many phrases
 * that have been defined for "validate_xdr_type" in a very similar
 * way.
 * TODO: Currently many checks have been left out
 *)


let rec expanded_xdr_type_term (s:xdr_type_term_system) (t:xdr_type_term) 
        : xdr_type_term =
  match t with
    X_array_fixed (t',n) ->
      X_array_fixed ((expanded_xdr_type_term s t'), n)
  | X_array (t',n) ->
      X_array ((expanded_xdr_type_term s t'), n)
  | X_struct st ->
      let s_names, s_types = List.split st in
      X_struct
	(List.combine
	   s_names
	   (List.map (expanded_xdr_type_term s) s_types))
  | X_union_over_int (u,default) ->
      let u_values, u_types = List.split u in
      let default' =
	match default with
	  Some d -> Some (expanded_xdr_type_term s d)
	| None   -> None
      in
      X_union_over_int
	(List.combine
	   u_values 
	   (List.map (expanded_xdr_type_term s) u_types), default')
  | X_union_over_uint (u,default) ->
      let u_values, u_types = List.split u in
      let default' =
	match default with
	  Some d -> Some (expanded_xdr_type_term s d)
	| None   -> None
      in
      X_union_over_uint
	(List.combine
	   u_values 
	   (List.map (expanded_xdr_type_term s) u_types), default')
  | X_union_over_enum (e,u,default) ->
      let u_values, u_types = List.split u in
      let default' =
	match default with
	  Some d -> Some (expanded_xdr_type_term s d)
	| None   -> None
      in
      X_union_over_enum
	( (expanded_xdr_type_term s e),
	 (List.combine
	    u_values 
	    (List.map (expanded_xdr_type_term s) u_types)),
	 default')
  | X_type n ->
      let rec r s1 s2 =
	match s2 with
	  []       -> failwith ("expanded_xdr_type_term: cannot resolve X_type " ^ n)
	| (n',t') :: s2' -> 
	      if n = n' then
		expanded_xdr_type_term s1 t'
	      else
		r (s1 @ [n',t']) s2'
      in
      r [] s
  | X_rec (n, t') ->
      X_rec (n, expanded_xdr_type_term s t')
  | _ ->
      t
;;


let expanded_xdr_type (s:xdr_type_system) (t:xdr_type_term) : xdr_type =
  try
    validate_xdr_type_i1 (expand_X_type s) [] t
  with
    Not_found -> failwith "expanded_xdr_type: unspecified error"
  | Propagate s -> failwith ("expanded_xdr_type: " ^ s)
;;


(**********************************************************************)
(* test on compatibility                                              *)
(**********************************************************************)

let are_compatible (s1:xdr_type) (s2:xdr_type) : bool = 
  (* implementation:
   * enum, struct and union members can be swapped
   *)

  failwith "are_compatible: not implemented"

;;


(**********************************************************************)
(* common implementation of value_matches_type & pack_xdr_value       *)
(**********************************************************************)

let rec pack 
        (v:xdr_value)
        (t:xdr_type)
        (get_param:string->xdr_type) 
        (print:string->unit) 
      : unit =

  let print_int4 x =
    print (int4_as_string x)
  in

  let print_uint4 x =
    print (uint4_as_string x)
  in

  let print_int8 x =
    print (int8_as_string x)
  in

  let print_uint8 x =
    print (uint8_as_string x)
  in

  let print_fp4 x =
    print (fp4_as_string x)
  in

  let print_fp8 x =
    print (fp8_as_string x)
  in

  let print_string s =
    let l = String.length s in
    let n = 4-(l mod 4) in
    if n < 4 then
      print (s ^ String.make n (Char.chr 0))
    else
      print s
  in

  match t.term with
    T_int ->
      let x = dest_xv_int v in
      print_int4 x 
  | T_uint ->
      let x = dest_xv_uint v in
      print_uint4 x 
  | T_hyper ->
      let x = dest_xv_hyper v in
      print_int8 x
  | T_uhyper ->
      let x = dest_xv_uhyper v in
      print_uint8 x 
  | T_enum e ->
      let x = dest_xv_enum v in
      let i = List.assoc x e in
      print_int4 i
  | T_float ->
      let x = dest_xv_float v in
      print_fp4 x 
  | T_double ->
      let x = dest_xv_double v in
      print_fp8 x
  | T_opaque_fixed n ->
      let x = dest_xv_opaque v in
      if String.length x = int_of_uint4 n then begin
	print_string x
      end
      else
	raise Not_found
  | T_opaque n ->
      let m = int_of_uint4 n in
      let x = dest_xv_opaque v in
      if String.length x <= m then begin
 	print_uint4 (uint4_of_int (String.length x));
	print_string x 
      end
      else
	raise Not_found
  | T_string n ->
      let m = int_of_uint4 n in
      let x = dest_xv_string v in
      if String.length x <= m then begin
 	  print_uint4 (uint4_of_int (String.length x));
	  print_string x 
      end
      else
	raise Not_found
  | T_array_fixed (t',n) ->
      let m = int_of_uint4 n in
      let x = dest_xv_array v in
      if List.length x = m then begin
	List.iter
	  (fun v' -> pack v' t' get_param print)
	  x
      end
      else
	raise Not_found
  | T_array (t',n) ->
      let m = int_of_uint4 n in
      let x = dest_xv_array v in
      if List.length x <= m then begin
	print_uint4 (uint4_of_int (List.length x));
	List.iter
	  (fun v' -> pack v' t' get_param print)
	  x
      end
      else
	raise Not_found
  | T_struct s ->
      let x = dest_xv_struct v in
      List.iter
	(fun (name,v') ->
	  ignore (List.assoc name s); ())
	x;
      List.iter
	(fun (name,t') ->
	  let v' = List.assoc name x in
	  pack v' t' get_param print)
	s
  | T_union_over_int (u,default) ->
      let i,x = dest_xv_union_over_int v in
      let t' =
      	try
	  List.assoc i u
	with
	  Not_found ->
	    match default with
	      Some d -> d
	    | None   -> raise Not_found
      in
      print_int4 i;
      pack x t' get_param print
  | T_union_over_uint (u,default) ->
      let i,x = dest_xv_union_over_uint v in
      let t' =
      	try
	  List.assoc i u
	with
	  Not_found ->
	    match default with
	      Some d -> d
	    | None   -> raise Not_found
      in
      print_uint4 i;
      pack x t' get_param print
  | T_union_over_enum (et,u,default) ->
      let s,x = dest_xv_union_over_enum v in
      let i =
	match et.term with
	  T_enum el -> List.assoc s el
	| _         -> failwith "pack: internal error"
      in
      let t' =
      	try
	  List.assoc s u
	with
	  Not_found ->
	    match default with
	      Some d -> d
	    | None   -> raise Not_found
      in
      print_int4 i;
      pack x t' get_param print
  | T_void ->
      ()
  | T_param n ->
      let t' = get_param n in
      pack v t' (fun n'->raise Not_found) print
  | T_rec (n, t') ->
      pack v t' get_param print
  | T_refer (n, t') ->
      pack v t' get_param print
;;


let value_matches_type 
    (v:xdr_value) 
    (t:xdr_type) 
    (p:(string * xdr_type) list)
  : bool =
  if Listext.equal_sets t.params (List.map fst p) &
     List.for_all (fun (n,t') -> t'.params = []) p then
    try
      pack v t (fun n -> List.assoc n p) (fun s -> ());
      true
    with
      _ ->      (* we assume here that no other errors can occur *)
      	false
  else
    false
;;


(**********************************************************************)
(* pack and unpack values                                             *)
(**********************************************************************)

let pack_xdr_value 
    (v:xdr_value) 
    (t:xdr_type)
    (p:(string * xdr_type) list)
    (print:string->unit)
  : unit =

  (* DEBUG *)
  (* List.iter (fun pn -> prerr_endline ("param " ^ pn)) t.params; *)

  if Listext.equal_sets t.params (List.map fst p) &
     List.for_all (fun (n,t') -> t'.params = []) p then
    try
      pack v t (fun n -> List.assoc n p) print
    with
      any ->
	(* DEBUG *)
	  prerr_endline (Printexc.to_string any);
      	failwith "pack_xdr_value"
  else
    failwith "pack_xdr_value"
;;


let pack_xdr_value_as_string
    (v:xdr_value) 
    (t:xdr_type)
    (p:(string * xdr_type) list)
  : string =
  let buf = Buffer.create 8192 in
  pack_xdr_value v t p (Buffer.add_string buf);
  Buffer.contents buf
;;


let unpack_term
    (str:string) 
    (t:xdr_type) 
    (get_param:string->xdr_type)
  : xdr_value =

  let strlen = String.length str in
  let k = ref 0 in

  let read_uint4 () =
    if !k + 4 <= strlen then begin
      k := !k + 4;
      mk_uint4 (str.[ !k - 4 ], str.[ !k - 3 ], str.[ !k - 2 ], str.[ !k - 1 ])
    end
    else
      raise (Xdr_format "message too short")
  in

  let read_int4 () =
    if !k + 4 <= strlen then begin
      k := !k + 4;
      mk_int4 (str.[ !k - 4 ], str.[ !k - 3 ], str.[ !k - 2 ], str.[ !k - 1 ])
    end
    else
      raise (Xdr_format "message too short")
  in

  let read_uint8 () =
    if !k + 8 <= strlen then begin
      k := !k + 8;
      mk_uint8 (str.[ !k - 8 ], str.[ !k - 7 ], str.[ !k - 6 ], str.[ !k - 5 ],
                str.[ !k - 4 ], str.[ !k - 3 ], str.[ !k - 2 ], str.[ !k - 1 ])
    end
    else
      raise (Xdr_format "message too short")
  in

  let read_int8 () =
    if !k + 8 <= strlen then begin
      k := !k + 8;
      mk_int8 (str.[ !k - 8 ], str.[ !k - 7 ], str.[ !k - 6 ], str.[ !k - 5 ],
               str.[ !k - 4 ], str.[ !k - 3 ], str.[ !k - 2 ], str.[ !k - 1 ])
    end
    else
      raise (Xdr_format "message too short")
  in

  let read_fp4 () =
    if !k + 4 <= strlen then begin
      k := !k + 4;
      mk_fp4 (str.[ !k - 4 ], str.[ !k - 3 ], str.[ !k - 2 ], str.[ !k - 1 ])
    end
    else
      raise (Xdr_format "message too short")
  in

  let read_fp8 () =
    if !k + 8 <= strlen then begin
      k := !k + 8;
      mk_fp8 (str.[ !k - 8 ], str.[ !k - 7 ], str.[ !k - 6 ], str.[ !k - 5 ],
              str.[ !k - 4 ], str.[ !k - 3 ], str.[ !k - 2 ], str.[ !k - 1 ])
    end
    else
      raise (Xdr_format "message too short")
  in

  let read_string n =
    let m = if n mod 4 = 0 then n else n+4-(n mod 4) in
    if !k + m <= strlen then begin
      k := !k + m;
      String.sub str (!k - m) n
    end
    else
      raise (Xdr_format "message too short")
  in

  let rec unpack t =
    match t.term with
      T_int ->
	XV_int (read_int4 ())
    | T_uint ->
	XV_uint (read_uint4 ())
    | T_hyper ->
	XV_hyper (read_int8 ())
    | T_uhyper ->
	XV_uhyper (read_uint8 ())
    | T_enum e ->
	let i = read_int4() in
	begin
	  try
	    XV_enum (Listext.rev_assoc i e)
	  with
	    Not_found -> raise (Xdr_format "value not included in enumeration")
	end
    | T_float ->
	XV_float (read_fp4 ())
    | T_double ->
	XV_double (read_fp8())
    | T_opaque_fixed n ->
	XV_opaque (read_string (int_of_uint4 n)) 
    | T_opaque n ->
	let m = read_uint4() in
	if int_of_uint4 m <= int_of_uint4 n then
	  XV_opaque (read_string (int_of_uint4 m))
	else
	  raise (Xdr_format "maximum length of field exceeded")
    | T_string n ->
	let m = read_uint4() in
	if int_of_uint4 m <= int_of_uint4 n then
	  XV_string (read_string (int_of_uint4 m))
	else
	  raise (Xdr_format "maximum length of field exceeded")
    | T_array_fixed (t',n) ->
	XV_array
	 (List.map
	    (fun d -> unpack t')
	    (Listext.upto 1 (int_of_uint4 n)))
    | T_array (t',n) ->
	let m = read_uint4() in
	if int_of_uint4 m <= int_of_uint4 n then begin
	  XV_array
	    (List.map
	       (fun d -> unpack t')
	       (Listext.upto 1 (int_of_uint4 m)))
	end
	else
	  raise (Xdr_format "maximum length of array exceeded")
    | T_struct s ->
	XV_struct 
	  (List.map
	     (fun (name,t') -> (name,unpack t'))
	     s)
    | T_union_over_int (u,default) ->
	let n = read_int4() in
	let t' =
	  try
	    List.assoc n u
	  with
	    Not_found ->
	      match default with
		None   -> raise (Xdr_format "undefined discriminator")
	      |	Some d -> d
	in
	XV_union_over_int (n, unpack t')
    | T_union_over_uint (u,default) ->
	let n = read_uint4() in
	let t' =
	  try
	    List.assoc n u
	  with
	    Not_found ->
	      match default with
		None   -> raise (Xdr_format "undefined discriminator")
	      |	Some d -> d
	in
	XV_union_over_uint (n, unpack t')
    | T_union_over_enum (e_type,u,default) ->
        let T_enum e = e_type.term in
	let n = read_int4() in
        let n_enum =
	  try
	    Listext.rev_assoc n e
	  with
	    Not_found -> raise (Xdr_format "value not included in enumeration")
	in
	let t' =
	  try
	    List.assoc n_enum u
	  with
	    Not_found ->
	      match default with
		None   -> raise (Xdr_format "undefined discriminator")
	      |	Some d -> d
	in
	XV_union_over_enum (n_enum, unpack t')
    | T_void ->
	XV_void
    | T_param p ->
	let t' = get_param p in
	unpack t'
    | T_rec (n, t') ->
	unpack t'
    | T_refer (n, t') ->
	unpack t'
  in
  try
    let v = unpack t in
    if !k = strlen then
      v
    else
      raise (Xdr_format_message_too_long v)
  with
    Cannot_represent ->
      raise (Xdr_format "implementation restriction")
;;


let unpack_xdr_value 
    (str:string) 
    (t:xdr_type) 
    (p:(string * xdr_type) list)
  : xdr_value =

  if Listext.equal_sets t.params (List.map fst p) &
     List.for_all (fun (n,t') -> t'.params = []) p then

    unpack_term str t (fun n -> List.assoc n p)

  else
    failwith "unpack_xdr_value"
;;

(* ======================================================================
 *
 * $Log: xdr.ml,v $
 * Revision 1.3  1999/04/11 22:16:12  gerd
 * 	Minor changes.
 *
 * Revision 1.2  1999/04/07 17:15:16  gerd
 * 	Improved quality of exceptions: There is now an explanation why
 * a type is illegal.
 *
 * Revision 1.1  1999/04/04 20:04:56  gerd
 * 	Initial release in this depot. This files have been restructured;
 * some missing features have been added (fp8 code, recursive xdr types).
 *
 *)
